/*
 * gcc -o getgroups getgroups.o
 */

#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <pwd.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <grp.h>
#include <malloc.h>

#include <crypt.h>
#include <shadow.h>

#define debugMessage printf

int unixGetGroups(char *name, gid_t basegid,
                  gid_t **groupList, int *ngroups,
                  int ignoreCase )

{
    gid_t *groups;
    struct group *gp;
    int maxSize;
    int count;
    int i;
    char **groupName;


    /* get groups, dynamically allocate as we go */

    maxSize=50;
    groups=malloc( sizeof( gid_t ) * maxSize );
    if ( groups==NULL )
        return -1;
    *groups=basegid;
    count=1;

    setgrent();
    debugMessage("calling getgrent\n");
    for ( gp=getgrent(); gp!=NULL; gp=getgrent() ) {
        /* skip this if already there */
        for ( i=0; i<count; i++ )
            if ( groups[i]==gp->gr_gid )
                break;
        if ( i<count )
            continue;

        debugMessage("Group: %s: ", gp->gr_name);
        for ( groupName=gp->gr_mem; *groupName!=NULL; groupName++ )
        {
            debugMessage("%s ", *groupName);
            if ( (ignoreCase==0 && strcmp( name,*groupName )==0) ||
                (ignoreCase!=0 && strcasecmp( name,*groupName )==0) )
            {
                if ( count+1==maxSize ) {
                    gid_t *newList;

                    newList=malloc( sizeof( gid_t ) * (maxSize+50) );
                    if ( newList==NULL ) {
                        endgrent();
                        free( groups );
                        return -1;
                    }
                    memcpy( newList,groups,sizeof(gid_t)*maxSize );
                    maxSize+=50;
                    free( groups );
                    groups=newList;
                }

                groups[count]=gp->gr_gid;
                count++;
            }
        }
        debugMessage("\n");
    }

    endgrent();

    *groupList=groups;
    *ngroups=count;

    debugMessage( "User %s in %d groups.\n",name,count );

    return 0;
}

int main(int argc, char **argv)
{
  int rc;
  struct passwd *pwent;
  gid_t *groupList;
  int ngroups = 0;
  int ignoreCase = 0;
  int i;

  pwent = getpwnam(argv[1]);
  if ( pwent==NULL ) {
    debugMessage( "User not in /etc/passwd.\n" );
    return -1;
  }

  if (argc == 3 && *argv[2] == 'T') {
    ignoreCase = 1;
    debugMessage("Ignoring case for group lookup\n");
  }

  debugMessage("%s:%s:%d:%d:%s:%s:%s\n", pwent->pw_name, pwent->pw_passwd, pwent->pw_uid, pwent->pw_gid, pwent->pw_gecos, pwent->pw_dir, pwent->pw_shell);
  rc = unixGetGroups(argv[1], pwent->pw_uid, &groupList, &ngroups, ignoreCase);
  if (ngroups) {
    debugMessage("Groups: ");
    for (i = 0; i < ngroups; i++) {
      debugMessage("%d ",  groupList[i]);
    }
    debugMessage("\n");
  } else {
    debugMessage("No groups\n");
  }

  debugMessage("calling getgroups\n");
  ngroups = getgroups(0, groupList);
  debugMessage("number of groups: %d\n", ngroups);
  groupList = malloc(ngroups*sizeof(gid_t));

  ngroups = getgroups(ngroups, groupList);
  if (ngroups) {
    debugMessage("Groups: ");
    for (i = 0; i < ngroups; i++) {
      debugMessage("%d ",  groupList[i]);
    }
    debugMessage("\n");
  } else {
    debugMessage("No groups\n");
  }
}
